//
//  BonjourSocketServer.m
//  SampleMulticast
//
//  Created by yuwei on 2011/5/3.
//  Copyright 2011年 __MyCompanyName__. All rights reserved.
//

#import "BonjourSocketServer.h"
#import "SocketServerRequest.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#import <arpa/inet.h>
#include <CFNetwork/CFSocketStream.h>


#define SERVICETYPE @"_sampleM._tcp."
#define PORTNUMBER 9090

NSString * const BonjourServerErrorDomain = @"BonjourServerErrorDomain";

@implementation BonjourSocketServer

@synthesize netService;
@synthesize ipv4socket;
@synthesize connectionPool;

@synthesize delegate;
#pragma mark - object lifecycle
- (id) init {
    self = [super init];
    if (!self) {
        return nil;
    }
    connectionPool = [[NSMutableSet alloc] init];
    NSError *thisError = nil;
    
    [self setupServer: &thisError];
    LOGGING_FACILITY( thisError == nil, [thisError localizedDescription] );
    [thisError release];
    
    return self;
}

- (void) dealloc {
    [self teardown];
    [connectionPool release];
    [super dealloc];
}

#pragma mark - socket callback

static void socketServerAcceptCallBack(CFSocketRef socket, CFSocketCallBackType type, CFDataRef address, const void *data, void *info) {
    BonjourSocketServer *server = (BonjourSocketServer *)info;
    if (kCFSocketAcceptCallBack == type) { 
        // for an AcceptCallBack, the data parameter is a pointer to a CFSocketNativeHandle
        CFSocketNativeHandle nativeSocketHandle = *(CFSocketNativeHandle *)data;
        struct sockaddr_in peerAddress;
        socklen_t peerLen = sizeof(peerAddress);
        NSString * peer = nil;
		
        if (getpeername(nativeSocketHandle, (struct sockaddr *)&peerAddress, (socklen_t *)&peerLen) == 0) {
            peer = [NSString stringWithUTF8String:inet_ntoa(peerAddress.sin_addr)];
		} else {
			peer = @"Generic Peer";
		}
		
        CFReadStreamRef readStream = NULL;
		CFWriteStreamRef writeStream = NULL;
        CFStreamCreatePairWithSocket(kCFAllocatorDefault, nativeSocketHandle, &readStream, &writeStream);
		
        if (readStream && writeStream) {
            CFReadStreamSetProperty(readStream, kCFStreamPropertyShouldCloseNativeSocket, kCFBooleanTrue);
            CFWriteStreamSetProperty(writeStream, kCFStreamPropertyShouldCloseNativeSocket, kCFBooleanTrue);
            [server handleConnection:peer inputStream:(NSInputStream *)readStream outputStream:(NSOutputStream *)writeStream];
        } else {
            // on any failure, need to destroy the CFSocketNativeHandle 
            // since we are not going to use it any more
            close(nativeSocketHandle);
        }
        if (readStream) CFRelease(readStream);
        if (writeStream) CFRelease(writeStream);
    }
}

#pragma mark - socket handleConnection

- (void)handleConnection:(NSString *)peerName inputStream:(NSInputStream *)readStream outputStream:(NSOutputStream *)writeStream {
    
    LOGGING_FACILITY( peerName != nil, @"No peer name given for client." );
	LOGGING_FACILITY( readStream != nil && writeStream != nil, @"One or both streams are invalid." );
	
	if (peerName != nil && readStream != nil && writeStream != nil) {
		SocketServerRequest * newPeer = [[SocketServerRequest alloc] initWithInputStream:readStream 
																			outputStream:writeStream 
																					peer:peerName 
																				delegate:self];
                
		if (newPeer) {
			[newPeer runProtocol];
			[self.connectionPool addObject:newPeer];
		}
        
		[newPeer release];
	}


}

#pragma mark - SocketServerRequest delegate

- (NSData *)respondToClient:(NSString *)url{
    return [delegate respondToClient:url];
}

- (void)socketServerRequestDidFinish:(SocketServerRequest *)request {
    if (request) {
		[self.connectionPool removeObject:request];
	}
}
- (void)socketServerRequestDidReceiveError:(SocketServerRequest *)request {
    if (request) {
		[self.connectionPool removeObject:request];
	}
}

#pragma mark - server methods

- (uint16_t) setupSocketServer:(NSError **)error {
    struct sockaddr_in serverAddress;
	socklen_t nameLen = 0;
	nameLen = sizeof(serverAddress);
    
    CFSocketContext socketCtxt = {0, self, NULL, NULL, NULL};
    self.ipv4socket = CFSocketCreate(kCFAllocatorDefault, PF_INET, SOCK_STREAM, IPPROTO_TCP, kCFSocketAcceptCallBack, (CFSocketCallBack)&socketServerAcceptCallBack, &socketCtxt);
	
    if (!ipv4socket) {
        if (error) * error = [[NSError alloc] initWithDomain:BonjourServerErrorDomain code:kBonjourServerNoSocketsAvailable userInfo:nil];
        [self teardown];
        return 0;
    }
    
    int yes = 1;
    setsockopt(CFSocketGetNative(ipv4socket), SOL_SOCKET, SO_REUSEADDR, (void *)&yes, sizeof(yes));
    
    // set up the IPv4 endpoint; use port 0, so the kernel will choose an arbitrary port for us, which will be advertised using Bonjour
    memset(&serverAddress, 0, sizeof(serverAddress));
    serverAddress.sin_len = nameLen;
    serverAddress.sin_family = AF_INET;
    serverAddress.sin_port = 0;
    serverAddress.sin_addr.s_addr = htonl(INADDR_ANY);
    NSData * address4 = [NSData dataWithBytes:&serverAddress length:nameLen];
    
    if (kCFSocketSuccess != CFSocketSetAddress(ipv4socket, (CFDataRef)address4)) {
        if (error) *error = [[NSError alloc] initWithDomain:BonjourServerErrorDomain code:kBonjourServerCouldNotBindToIPv4Address userInfo:nil];
        if (ipv4socket) CFRelease(ipv4socket);
        ipv4socket = NULL;
        return 0;
    }
    
    // now that the binding was successful, we get the port number 
    // -- we will need it for the NSNetService
    NSData * addr = [(NSData *)CFSocketCopyAddress(ipv4socket) autorelease];
    memcpy(&serverAddress, [addr bytes], [addr length]);
    return ntohs(serverAddress.sin_port);
    
}

- (void) setupServer:(NSError **)error {
    uint16_t chosenPort;
    
    if (self.netService && ipv4socket) {
        return;
    }
    
    if (!ipv4socket) {
        chosenPort = [self setupSocketServer: error];
    }
    if (!ipv4socket) {
        return;
    }
    
    if (chosenPort == 0) {
        chosenPort = PORTNUMBER;
    }
    // set up the run loop sources for the sockets
    CFRunLoopRef cfrl = CFRunLoopGetCurrent();
    CFRunLoopSourceRef source = CFSocketCreateRunLoopSource(kCFAllocatorDefault, ipv4socket, 0);
    CFRunLoopAddSource(cfrl, source, kCFRunLoopCommonModes);
    CFRelease(source);
    
    if (!self.netService && ipv4socket) {
        NSString *serviceType = SERVICETYPE;
        self.netService = [[NSNetService alloc] initWithDomain:@"local" type: serviceType name:[[UIDevice currentDevice] name] port:chosenPort];
        [self.netService setDelegate:self];
    }
    
    if(!self.netService && !ipv4socket){
        if (error) *error = [[NSError alloc] initWithDomain:BonjourServerErrorDomain code:kBonjourServerCouldNotBindOrEstablishNetService userInfo:nil];
        [self teardown];
        return;
    }
}

- (void) run {
    NSError * thisError = nil;
	[self setupServer: &thisError];
	
	LOGGING_FACILITY( thisError == nil, [thisError localizedDescription] );
	[thisError release];
	
	[self.netService scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSRunLoopCommonModes];
	[self.netService publish];
}

- (void) teardown{
    if (self.netService) {
		[self.netService stop];
		[self.netService removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSRunLoopCommonModes];
		self.netService = nil;
	}
}

#pragma mark - NSNetService Delegate

- (void)netServiceDidPublish:(NSNetService *)sender {
    self.netService = sender;
}

#pragma mark - Bonjour TXT Reord

- (void) update {
    NSMutableString *message = [[delegate cacheList] retain];
    NSData *d = [message dataUsingEncoding:NSUTF8StringEncoding];
    NSDictionary *txtDict = [NSDictionary dictionaryWithObject:d forKey:@"message"];
    NSData *txtData = [NSNetService dataFromTXTRecordDictionary: txtDict];
    [netService setTXTRecordData: txtData];
    [message autorelease];
}

- (void) remove: (NSString *)message {
    NSData *d = [message dataUsingEncoding:NSUTF8StringEncoding];
    NSDictionary *txtDict = [NSDictionary dictionaryWithObject:d forKey:@"message"];
    NSData *txtData = [NSNetService dataFromTXTRecordDictionary: txtDict];
    [netService setTXTRecordData: txtData];
}

@end
